@@ -2,41 +2,34 @@ from __future__ import absolute_import
 
 import collections, copy, operator, re
 
-from .schema import SolrError, SolrUnicodeField, SolrBooleanField
-from .strings import WildcardString
+from .schema import SolrError, SolrBooleanField, SolrUnicodeField, WildcardFieldInstance
 
 
 class LuceneQuery(object):
     default_term_re = re.compile(r'^\w+$')
-    range_query_templates = {
-        "lt": "{* TO %s}",
-        "lte": "[* TO %s]",
-        "gt": "{%s TO *}",
-        "gte": "[%s TO *]",
-        "rangeexc": "{%s TO %s}",
-        "range": "[%s TO %s]",
-    }
     def __init__(self, schema, option_flag=None, original=None):
         self.schema = schema
+        self.normalized = False
         if original is None:
             self.option_flag = option_flag
             self.terms = collections.defaultdict(set)
             self.phrases = collections.defaultdict(set)
             self.ranges = set()
             self.subqueries = []
-            self._or = self._and = self._not = self._pow = None
+            self._and = True
+            self._or = self._not = self._pow = False
             self.boosts = []
         else:
             self.option_flag = original.option_flag
             self.terms = copy.copy(original.terms)
             self.phrases = copy.copy(original.phrases)
             self.ranges = copy.copy(original.ranges)
-            self.subqueries = [q.clone() for q in original.subqueries]
+            self.subqueries = copy.copy(original.subqueries)
             self._or = original._or
             self._and = original._and
             self._not = original._not
             self._pow = original._pow
-            self.boosts = original.boosts
+            self.boosts = copy.copy(original.boosts)
 
     def clone(self):
         return LuceneQuery(self.schema, original=self)
@@ -48,58 +41,145 @@ class LuceneQuery(object):
             opts[self.option_flag] = s
         return opts
 
+    def serialize_debug(self, indent=0):
+        indentspace = indent * ' '
+        print '%s%s (%s)' % (indentspace, repr(self), "Normalized" if self.normalized else "Not normalized")
+        print '%s%s' % (indentspace, '{')
+        for term in self.terms.items():
+            print '%s%s' % (indentspace, term)
+        for phrase in self.phrases.items():
+            print '%s%s' % (indentspace, phrase)
+        for range in self.ranges:
+            print '%s%s' % (indentspace, range)
+        if self.subqueries:
+            if self._and:
+                print '%sAND:' % indentspace
+            elif self._or:
+                print '%sOR:' % indentspace
+            elif self._not:
+                print '%sNOT:' % indentspace
+            elif self._pow is not False:
+                print '%sPOW %s:' % (indentspace, self._pow)
+            else:
+                raise ValueError
+            for subquery in self.subqueries:
+                subquery.serialize_debug(indent+2)
+        print '%s%s' % (indentspace, '}')
+
     # Below, we sort all our value_sets - this is for predictability when testing.
-    def serialize_term_queries(self):
+    def serialize_term_queries(self, terms):
         s = []
-        for name, value_set in sorted(self.terms.items()):
+        for name, value_set in terms.items():
             if name:
                 field = self.schema.match_field(name)
             else:
                 field = self.schema.default_field
-            if isinstance(field, SolrUnicodeField):
-                value_set = [value.escape_for_lqs_term() for value in value_set]
             if name:
-                s += [u'%s:%s' % (name, value) for value in sorted(value_set)]
+                s += [u'%s:%s' % (name, value.to_query()) for value in value_set]
             else:
-                s += sorted(value_set)
-        return ' AND '.join(s)
-
-    def serialize_phrase_queries(self):
-        s = []
-        for name, value_set in sorted(self.phrases.items()):
-            if name:
-                field = self.schema.match_field(name)
-            else:
-                field = self.schema.default_field
-            if isinstance(field, SolrUnicodeField):
-                value_set = [value.escape_for_lqs_term() for value in value_set]
-            if name:
-                s += [u'%s:"%s"' % (name, value)
-                      for value in sorted(value_set)]
-            else:
-                s += ['"%s"' % value for value in sorted(value_set)]
-        return ' AND '.join(s)
+                s += [value.to_query() for value in value_set]
+        return u' AND '.join(sorted(s))
 
+    range_query_templates = {
+        "any": u"[* TO *]",
+        "lt": u"{* TO %s}",
+        "lte": u"[* TO %s]",
+        "gt": u"{%s TO *}",
+        "gte": u"[%s TO *]",
+        "rangeexc": u"{%s TO %s}",
+        "range": u"[%s TO %s]",
+    }
     def serialize_range_queries(self):
         s = []
-        for name, rel, value in sorted(self.ranges):
-            range = self.range_query_templates[rel] % value
-            s.append("%(name)s:%(range)s" % vars())
-        return ' AND '.join(s)
+        for name, rel, values in sorted(self.ranges):
+            range_s = self.range_query_templates[rel] % \
+                tuple(value.to_query() for value in sorted(values, key=lambda x: getattr(x, "value")))
+            s.append(u"%s:%s" % (name, range_s))
+        return u' AND '.join(s)
 
-    def child_needs_parens(self, child, op=None):
+    def child_needs_parens(self, child):
         if len(child) == 1:
             return False
-        elif child._not is not None or child._pow is not None:
-            return False
-        elif (self._or is not None or op=='OR') and child._or is not None:
-            return False
-        elif (self._and is not None or op=='AND') and child._and is not None:
-            return False
+        elif self._or:
+            return not (child._or or child._pow)
+        elif (self._and or self._not):
+            return not (child._and or child._not or child._pow)
+        elif self._pow is not False:
+            return True
         else:
             return True
 
-    def __unicode__(self):
+    @staticmethod
+    def merge_term_dicts(*args):
+        d = collections.defaultdict(set)
+        for arg in args:
+            for k, v in arg.items():
+                d[k].update(v)
+        return dict((k, v) for k, v in d.items())
+
+    def normalize(self):
+        if self.normalized:
+            return self, False
+        mutated = False
+        _subqueries = []
+        _terms = self.terms
+        _phrases = self.phrases
+        _ranges = self.ranges
+        for s in self.subqueries:
+            _s, changed = s.normalize()
+            if not _s or changed:
+                mutated = True
+            if _s:
+                if (_s._and and self._and) or (_s._or and self._or):
+                    mutated = True
+                    _terms = self.merge_term_dicts(_terms, _s.terms)
+                    _phrases = self.merge_term_dicts(_phrases, _s.phrases)
+                    _ranges = _ranges.union(_s.ranges)
+                    _subqueries.extend(_s.subqueries)
+                else:
+                    _subqueries.append(_s)
+        if mutated:
+            newself = self.clone()
+            newself.terms = _terms
+            newself.phrases = _phrases
+            newself.ranges = _ranges
+            newself.subqueries = _subqueries
+            self = newself
+
+        if self._not:
+            if not len(self.subqueries):
+                newself = self.clone()
+                newself._not = False
+                newself._and = True
+                self = newself
+                mutated = True
+            elif len(self.subqueries) == 1:
+                if self.subqueries[0]._not:
+                    newself = self.clone()
+                    newself.subqueries = self.subqueries[0].subqueries
+                    newself._not = False
+                    newself._and = True
+                    self = newself
+                    mutated = True
+            else:
+                raise ValueError
+        elif self._pow:
+            if not len(self.subqueries):
+                newself = self.clone()
+                newself._pow = False
+                self = newself
+                mutated = True
+        elif self._and or self._or:
+            if not self.terms and not self.phrases and not self.ranges:
+                if len(self.subqueries) == 1:
+                    self = self.subqueries[0]
+                    mutated = True
+        self.normalized = True
+        return self, mutated
+
+    def __unicode__(self, level=0, op=None):
+        if not self.normalized:
+            self, _ = self.normalize()
         if self.boosts:
             # Clone and rewrite to effect the boosts.
             newself = self.clone()
@@ -107,61 +187,34 @@ class LuceneQuery(object):
             boost_queries = [self.Q(**kwargs)**boost_score
                              for kwargs, boost_score in self.boosts]
             newself = newself | (newself & reduce(operator.or_, boost_queries))
-            return unicode(newself)
-        if self._or is not None:
-            s = []
-            for o in self._or:
-                if self.child_needs_parens(o):
-                    s.append(u"(%s)"%o)
-                else:
-                    s.append(u"%s"%o)
-            return u" OR ".join(s)
-        elif self._and is not None:
-            s = []
-            for o in self._and:
-                if self.child_needs_parens(o):
-                    s.append(u"(%s)"%o)
-                else:
-                    s.append(u"%s"%o)
-            return u" AND ".join(s)
-        elif self._not is not None:
-            o = self._not
-            if o._not is not None:
-                # they cancel out
-                self._not = self._not._not
-                return u"%s"%o
-            if self.child_needs_parens(o):
-                return u"NOT (%s)"%o
-            else:
-                return u"NOT %s"%o
-        elif self._pow is not None:
-            q, v = self._pow
-            if self.child_needs_parens(q):
-                return u"(%s)^%s"%(q,v)
-            else:
-                return u"%s^%s"%(q,v)
+            newself, _ = newself.normalize()
+            return newself.__unicode__(level=level)
         else:
-            u = [s for s in [self.serialize_term_queries(),
-                             self.serialize_phrase_queries(),
+            u = [s for s in [self.serialize_term_queries(self.terms),
+                             self.serialize_term_queries(self.phrases),
                              self.serialize_range_queries()]
                  if s]
-            if not u and len(self.subqueries) == 1:
-                # Only one subquery, no need for parens
-                u.append(u"%s"%self.subqueries[0])
+            for q in self.subqueries:
+                op_ = u'OR' if self._or else u'AND'
+                if self.child_needs_parens(q):
+                    u.append(u"(%s)"%q.__unicode__(level=level+1, op=op_))
+                else:
+                    u.append(u"%s"%q.__unicode__(level=level+1, op=op_))
+            if self._and:
+                return u' AND '.join(u)
+            elif self._or:
+                return u' OR '.join(u)
+            elif self._not:
+                assert len(u) == 1
+                if level == 0 or (level == 1 and op == "AND"):
+                    return u'NOT %s'%u[0]
+                else:
+                    return u'(*:* AND NOT %s)'%u[0]
+            elif self._pow is not False:
+                assert len(u) == 1
+                return u"%s^%s"%(u[0], self._pow)
             else:
-                for q in self.subqueries:
-                    if self.child_needs_parens(q, 'AND'):
-                        u.append(u"(%s)"%q)
-                    else:
-                        u.append(u"%s"%q)
-            return ' AND '.join(u)
-
-    def stringify_with_optional_parens(self):
-        if len(self) == 1:
-            s = u"%s"
-        else:
-            s = u"(%s)"
-        return s % self
+                raise ValueError
 
     def __len__(self):
         # How many terms in this (sub) query?
@@ -169,7 +222,10 @@ class LuceneQuery(object):
             subquery_length = len(self.subqueries[0])
         else:
             subquery_length = len(self.subqueries)
-        return sum([len(self.terms), len(self.phrases), len(self.ranges), subquery_length])
+        return sum([sum(len(v) for v in self.terms.values()),
+                    sum(len(v) for v in self.phrases.values()),
+                    len(self.ranges),
+                    subquery_length])
 
     def Q(self, *args, **kwargs):
         q = LuceneQuery(self.schema)
@@ -181,17 +237,21 @@ class LuceneQuery(object):
 
     def __or__(self, other):
         q = LuceneQuery(self.schema)
-        q._or = (self, other)
+        q._and = False
+        q._or = True
+        q.subqueries = [self, other]
         return q
 
     def __and__(self, other):
         q = LuceneQuery(self.schema)
-        q._and = (self, other)
+        q.subqueries = [self, other]
         return q
 
     def __invert__(self):
         q = LuceneQuery(self.schema)
-        q._not = self
+        q._and = False
+        q._not = True
+        q.subqueries = [self]
         return q
 
     def __pow__(self, value):
@@ -200,10 +260,13 @@ class LuceneQuery(object):
         except ValueError:
             raise ValueError("Non-numeric value supplied for boost")
         q = LuceneQuery(self.schema)
-        q._pow = (self, value)
+        q.subqueries = [self]
+        q._and = False
+        q._pow = value
         return q
         
-    def add(self, args, kwargs, terms_or_phrases=None):
+    def add(self, args, kwargs):
+        self.normalized = False
         _args = []
         for arg in args:
             if isinstance(arg, LuceneQuery):
@@ -224,27 +287,40 @@ class LuceneQuery(object):
                 field_name, rel = k, 'eq'
             field = self.schema.match_field(field_name)
             if not field:
-                raise ValueError("%s is not a valid field name" % k)
+                if (k, v) != ("*", "*"):
+                    # the only case where wildcards in field names are allowed
+                    raise ValueError("%s is not a valid field name" % k)
+            elif not field.indexed:
+                raise SolrError("Can't query on non-indexed field '%s'" % field_name)
             if rel == 'eq':
                 self.add_exact(field_name, v, terms_or_phrases)
             else:
                 self.add_range(field_name, rel, v)
 
-    def add_exact(self, field_name, value, term_or_phrase):
-        if field_name:
-            field = self.schema.match_field(field_name)
-        else:
-            field = self.schema.default_field
-        values = field.serialize(value) # Might be multivalued
-        if isinstance(values, basestring):
+    def add_exact(self, field_name, values, term_or_phrase):
+        # We let people pass in a list of values to match.
+        # This really only makes sense for text fields or
+        # multivalued fields.
+        if not hasattr(values, "__iter__"):
             values = [values]
-        for value in values:
+        # We can only do a field_name == "*" if:
+        if field_name and field_name != "*":
+            field = self.schema.match_field(field_name)
+        elif not field_name:
+            field = self.schema.default_field
+        else: # field_name must be "*"
+            if len(values) == 1 and values[0] == "*":
+                self.terms["*"].add(WildcardFieldInstance.from_user_data())
+                return
+            else:
+                raise SolrError("If field_name is '*', then only '*' is permitted as the query")
+        insts = [field.instance_from_user_data(value) for value in values]
+        for inst in insts:
             if isinstance(field, SolrUnicodeField):
-                this_term_or_phrase = term_or_phrase or self.term_or_phrase(value)
-                value = WildcardString(value)
+                this_term_or_phrase = term_or_phrase or self.term_or_phrase(inst.value)
             else:
                 this_term_or_phrase = "terms"
-            getattr(self, this_term_or_phrase)[field_name].add(value)
+            getattr(self, this_term_or_phrase)[field_name].add(inst)
 
     def add_range(self, field_name, rel, value):
         field = self.schema.match_field(field_name)
@@ -258,10 +334,14 @@ class LuceneQuery(object):
             except (AssertionError, TypeError):
                 raise SolrError("'%s__%s' argument must be a length-2 iterable"
                                  % (field_name, rel))
-            value = tuple(sorted(field.serialize(v) for v in value))
+            insts = tuple(sorted(field.instance_from_user_data(v) for v in value))
+        elif rel == 'any':
+            if value is not True:
+                raise SolrError("'%s__%s' argument must be True")
+            insts = ()
         else:
-            value = field.serialize(value)
-        self.ranges.add((field_name, rel, value))
+            insts = (field.instance_from_user_data(value),)
+        self.ranges.add((field_name, rel, insts))
 
     def term_or_phrase(self, arg, force=None):
         return 'terms' if self.default_term_re.match(arg) else 'phrases'
@@ -269,23 +349,28 @@ class LuceneQuery(object):
     def add_boost(self, kwargs, boost_score):
         for k, v in kwargs.items():
             field = self.schema.match_field(k)
-            value = field.serialize(v)
+            if not field:
+                raise ValueError("%s is not a valid field name" % k)
+            elif not field.indexed:
+                raise SolrError("Can't query on non-indexed field '%s'" % field_name)
+            value = field.instance_from_user_data(v)
         self.boosts.append((kwargs, boost_score))
 
 
 class SolrSearch(object):
-    option_modules = ('query_obj', 'filter_obj', 'paginator', 'more_like_this', 'highlighter', 'faceter', 'sorter', 'facet_querier')
+    option_modules = ('query_obj', 'filter_obj', 'paginator', 'more_like_this', 'highlighter', 'faceter', 'sorter', 'facet_querier', 'field_limiter',)
     def __init__(self, interface, original=None):
         self.interface = interface
         self.schema = interface.schema
         if original is None:
-            self.query_obj = LuceneQuery(self.schema, 'q')
-            self.filter_obj = LuceneQuery(self.schema, 'fq')
+            self.query_obj = LuceneQuery(self.schema, u'q')
+            self.filter_obj = LuceneQuery(self.schema, u'fq')
             self.paginator = PaginateOptions(self.schema)
             self.more_like_this = MoreLikeThisOptions(self.schema)
             self.highlighter = HighlightOptions(self.schema)
             self.faceter = FacetOptions(self.schema)
             self.sorter = SortOptions(self.schema)
+            self.field_limiter = FieldLimitOptions(self.schema)
             self.facet_querier = FacetQueryOptions(self.schema)
         else:
             for opt in self.option_modules:
@@ -359,6 +444,11 @@ class SolrSearch(object):
         newself.sorter.update(field)
         return newself
 
+    def field_limit(self, fields=None, score=False, all_fields=False):
+        newself = self.clone()
+        newself.field_limiter.update(fields, score, all_fields)
+        return newself
+
     def boost_relevancy(self, boost_score, **kwargs):
         if not self.query_obj:
             raise TypeError("Can't boost the relevancy of an empty query")
@@ -367,7 +457,6 @@ class SolrSearch(object):
         except ValueError:
             raise ValueError("Non-numeric boost value supplied")
 
-        # Clone all of self *except* query, which we'll take care of directly
         newself = self.clone()
         newself.query_obj.add_boost(kwargs, boost_score)
         return newself
@@ -376,12 +465,13 @@ class SolrSearch(object):
         options = {}
         for option_module in self.option_modules:
             options.update(getattr(self, option_module).options())
-        if 'q' not in options:
-            options['q'] = '*' # search everything
-        return options
+        if u'q' not in options:
+            options[u'q'] = u'*:*' # search everything
+        # Next line is for pre-2.6.5 python
+        return dict((k.encode('utf8'), v) for k, v in options.items())
 
     def params(self):
-        return self.interface.params_from_dict(**self.options())
+        return params_from_dict(**self.options())
 
     def execute(self, constructor=dict):
         result = self.interface.search(**self.options())
@@ -633,6 +723,43 @@ class SortOptions(Options):
             return {}
 
 
+class FieldLimitOptions(Options):
+    option_name = "fl"
+
+    def __init__(self, schema, original=None):
+        self.schema = schema
+        if original is None:
+            self.fields = set()
+            self.score = False
+            self.all_fields = False
+        else:
+            self.fields = copy.copy(original.fields)
+            self.score = original.score
+            self.all_fields = original.all_fields
+
+    def update(self, fields=None, score=False, all_fields=False):
+        if fields is None:
+            fields = []
+        if isinstance(fields, basestring):
+            fields = [fields]
+        self.schema.check_fields(fields, {"stored": True})
+        self.fields.update(fields)
+        self.score = score
+        self.all_fields = all_fields
+
+    def options(self):
+        opts = {}
+        if self.all_fields:
+            fields = set("*")
+        else:
+            fields = self.fields
+        if self.score:
+            fields.add("score")
+        if fields:
+            opts['fl'] = ','.join(sorted(fields))
+        return opts
+
+
 class FacetQueryOptions(Options):
     def __init__(self, schema, original=None):
         self.schema = schema
@@ -650,3 +777,20 @@ class FacetQueryOptions(Options):
                     'facet':True}
         else:
             return {}
+
+def params_from_dict(**kwargs):
+    utf8_params = []
+    for k, vs in kwargs.items():
+        if isinstance(k, unicode):
+            k = k.encode('utf-8')
+        # We allow for multivalued options with lists.
+        if not hasattr(vs, "__iter__"):
+            vs = [vs]
+        for v in vs:
+            if isinstance(v, bool):
+                v = u"true" if v else u"false"
+            else:
+                v = unicode(v)
+            v = v.encode('utf-8')
+            utf8_params.append((k, v))
+    return sorted(utf8_params)
