@@ -9,7 +9,10 @@ import datetime
 import mx.DateTime
 
 from .schema import SolrSchema, SolrError
-from .search import SolrSearch, PaginateOptions, FacetOptions, HighlightOptions, MoreLikeThisOptions
+from .search import SolrSearch, PaginateOptions, SortOptions, FieldLimitOptions, FacetOptions, HighlightOptions, MoreLikeThisOptions, params_from_dict
+from .strings import RawString
+
+debug = False
 
 schema_string = \
 """<schema name="timetric" version="1.1">
@@ -51,8 +54,6 @@ schema = SolrSchema(StringIO(schema_string))
 
 class MockInterface(object):
     schema = schema
-    def search(self, **kwargs):
-        return kwargs
 
 
 interface = MockInterface()
@@ -61,76 +62,74 @@ interface = MockInterface()
 good_query_data = {
     "query_by_term":(
         (["hello"], {},
-         {"q":u"hello"}),
+         [("q", u"hello")]),
         (["hello"], {"int_field":3},
-         {"q":u"hello int_field:3"}),
+         [("q", u"hello AND int_field:3")]),
         (["hello", "world"], {},
-         {"q":u"hello world"}),
+         [("q", u"hello AND world")]),
         # NB this next is not really what we want,
         # probably this should warn
         (["hello world"], {},
-         {"q":u"hello world"}),
+         [("q", u"hello\\ world")]),
         ),
 
     "query_by_phrase":(
         (["hello"], {},
-         # Do we actually want this many quotes in here?
-         {"q":u"\"hello\""}),
+         [("q", u"hello")]),
         (["hello"], {"int_field":3},
-         {"q":u"int_field:3 \"hello\""}), # Non-text data is always taken to be a term, and terms come before phrases, so order is reversed
+         [("q", u"int_field:3 AND hello")]), # Non-text data is always taken to be a term, and terms come before phrases, so order is reversed
         (["hello", "world"], {},
-         {"q":u"\"hello\" \"world\""}),
+         [("q", u"hello AND world")]),
         (["hello world"], {},
-         {"q":u"\"hello world\""}),
+         [("q", u"hello\\ world")]),
         ([], {'string_field':['hello world', 'goodbye, cruel world']},
-         {"q":u"string_field:\"goodbye, cruel world\" string_field:\"hello world\""}),
+         [("q", u"string_field:goodbye,\\ cruel\\ world AND string_field:hello\\ world")]),
         ),
 
     "filter_by_term":(
         (["hello"], {},
-         {"fq":u"hello"}),
+         [("fq", u"hello"), ("q", "*:*")]),
         (["hello"], {"int_field":3},
-         {"fq":u"hello int_field:3"}),
+         [("fq", u"hello AND int_field:3"), ("q", "*:*")]),
         (["hello", "world"], {},
-         {"fq":u"hello world"}),
+         [("fq", u"hello AND world"), ("q", "*:*")]),
         # NB this next is not really what we want,
         # probably this should warn
         (["hello world"], {},
-         {"fq":u"hello world"}),
+         [("fq", u"hello\\ world"), ("q", "*:*")]),
         ),
 
     "filter_by_phrase":(
         (["hello"], {},
-         # Do we actually want this many quotes in here?
-         {"fq":u"\"hello\""}),
+         [("fq", u"hello"), ("q", "*:*")]),
         (["hello"], {"int_field":3},
-         {"fq":u"int_field:3 \"hello\""}),
+         [("fq", u"int_field:3 AND hello"), ("q", "*:*")]),
         (["hello", "world"], {},
-         {"fq":u"\"hello\" \"world\""}),
+         [("fq", u"hello AND world"), ("q", "*:*")]),
         (["hello world"], {},
-         {"fq":u"\"hello world\""}),
+         [("fq", u"hello\\ world"), ("q", "*:*")]),
         ),
 
     "query":(
         (["hello"], {},
-         {"q":u"hello"}),
+         [("q", u"hello")]),
         (["hello"], {"int_field":3},
-         {"q":u"hello int_field:3"}),
+         [("q", u"hello AND int_field:3")]),
         (["hello", "world"], {},
-         {"q":u"hello world"}),
+         [("q", u"hello AND world")]),
         (["hello world"], {},
          {"q":u"\"hello world\""}),
         ),
 
     "filter":(
         (["hello"], {},
-         {"fq":u"hello"}),
+         [("fq", u"hello"), ("q", "*:*")]),
         (["hello"], {"int_field":3},
-         {"fq":u"hello int_field:3"}),
+         [("fq", u"hello AND int_field:3"), ("q", "*:*")]),
         (["hello", "world"], {},
-         {"fq":u"hello world"}),
+         [("fq", u"hello AND world"), ("q", "*:*")]),
         (["hello world"], {},
-         {"fq":u"\"hello world\""}),
+         [("fq", u"hello\\ world"), ("q", "*:*")]),
         ),
 
     "query":(
@@ -169,49 +168,63 @@ good_query_data = {
         ([], {"sdouble_field":3}, # casting from int should work
          {"q":u"sdouble_field:3.0"}),
         ([], {"date_field":datetime.datetime(2009, 1, 1)},
-         {"q":u"date_field:2009-01-01T00:00:00.000000Z"}),
+         {"q":u"date_field:2009-01-01T00\\:00\\:00.000000Z"}),
         ([], {"date_field":mx.DateTime.DateTime(2009, 1, 1)},
-         {"q":u"date_field:2009-01-01T00:00:00.000000Z"}),
+         {"q":u"date_field:2009-01-01T00\\:00\\:00.000000Z"}),
         ),
 
     "query":(
+        ([], {"int_field__any":True},
+         [("q", u"int_field:[* TO *]")]),
         ([], {"int_field__lt":3},
-         {"q":u"int_field:{* TO 3}"}),
+         [("q", u"int_field:{* TO 3}")]),
         ([], {"int_field__gt":3},
-         {"q":u"int_field:{3 TO *}"}),
+         [("q", u"int_field:{3 TO *}")]),
         ([], {"int_field__rangeexc":(-3, 3)},
-         {"q":u"int_field:{-3 TO 3}"}),
+         [("q", u"int_field:{\-3 TO 3}")]),
         ([], {"int_field__rangeexc":(3, -3)},
-         {"q":u"int_field:{-3 TO 3}"}),
+         [("q", u"int_field:{\-3 TO 3}")]),
         ([], {"int_field__lte":3},
-         {"q":u"int_field:[* TO 3]"}),
+         [("q", u"int_field:[* TO 3]")]),
         ([], {"int_field__gte":3},
-         {"q":u"int_field:[3 TO *]"}),
+         [("q", u"int_field:[3 TO *]")]),
         ([], {"int_field__range":(-3, 3)},
-         {"q":u"int_field:[-3 TO 3]"}),
+         [("q", u"int_field:[\-3 TO 3]")]),
         ([], {"int_field__range":(3, -3)},
-         {"q":u"int_field:[-3 TO 3]"}),
+         [("q", u"int_field:[\-3 TO 3]")]),
         ([], {"date_field__lt":datetime.datetime(2009, 1, 1)},
-         {"q":u"date_field:{* TO 2009-01-01T00:00:00.000000Z}"}),
+         [("q", u"date_field:{* TO 2009\\-01\\-01T00\\:00\\:00.000000Z}")]),
         ([], {"date_field__gt":datetime.datetime(2009, 1, 1)},
-         {"q":u"date_field:{2009-01-01T00:00:00.000000Z TO *}"}),
+         [("q", u"date_field:{2009\\-01\\-01T00\\:00\\:00.000000Z TO *}")]),
         ([], {"date_field__rangeexc":(datetime.datetime(2009, 1, 1), datetime.datetime(2009, 1, 2))},
-         {"q":u"date_field:{2009-01-01T00:00:00.000000Z TO 2009-01-02T00:00:00.000000Z}"}),
+         [("q", "date_field:{2009\\-01\\-01T00\\:00\\:00.000000Z TO 2009\\-01\\-02T00\\:00\\:00.000000Z}")]),
         ([], {"date_field__lte":datetime.datetime(2009, 1, 1)},
-         {"q":u"date_field:[* TO 2009-01-01T00:00:00.000000Z]"}),
+         [("q", u"date_field:[* TO 2009\\-01\\-01T00\\:00\\:00.000000Z]")]),
         ([], {"date_field__gte":datetime.datetime(2009, 1, 1)},
-         {"q":u"date_field:[2009-01-01T00:00:00.000000Z TO *]"}),
+         [("q", u"date_field:[2009\\-01\\-01T00\\:00\\:00.000000Z TO *]")]),
         ([], {"date_field__range":(datetime.datetime(2009, 1, 1), datetime.datetime(2009, 1, 2))},
-         {"q":u"date_field:[2009-01-01T00:00:00.000000Z TO 2009-01-02T00:00:00.000000Z]"}),
+         [("q", u"date_field:[2009\\-01\\-01T00\\:00\\:00.000000Z TO 2009\\-01\\-02T00\\:00\\:00.000000Z]")]),
         ([], {'string_field':['hello world', 'goodbye, cruel world']},
-         {"q":u"string_field:\"goodbye, cruel world\" string_field:\"hello world\""}),
-
+         [("q", u"string_field:goodbye,\\ cruel\\ world AND string_field:hello\\ world")]),
+        # Raw strings
+        ([], {'string_field':RawString("abc*???")},
+         [("q", "string_field:abc\\*\\?\\?\\?")]),
         ),
     }
 
 def check_query_data(method, args, kwargs, output):
     solr_search = SolrSearch(interface)
-    assert getattr(solr_search, method)(*args, **kwargs).execute() == output
+    p = getattr(solr_search, method)(*args, **kwargs).params()
+    try:
+        assert p == output
+    except AssertionError:
+        if debug:
+            print p
+            print output
+            import pdb;pdb.set_trace()
+            raise
+        else:
+            raise
 
 def test_query_data():
     for method, data in good_query_data.items():
@@ -241,7 +254,7 @@ bad_query_data = (
 def check_bad_query_data(kwargs):
     solr_search = SolrSearch(interface)
     try:
-        solr_search.query(**kwargs).execute()
+        solr_search.query(**kwargs).params()
     except SolrError:
         pass
     else:
@@ -277,6 +290,12 @@ good_option_data = {
         ({"fields":["int_field", "text_field"], "prefix":"abc", "limit":3},
          {"facet":True, "facet.field":["int_field", "text_field"], "f.int_field.facet.prefix":"abc", "f.int_field.facet.limit":3, "f.text_field.facet.prefix":"abc", "f.text_field.facet.limit":3, }),
         ),
+    SortOptions:(
+        ({"field":"int_field"},
+         {"sort":"int_field asc"}),
+        ({"field":"-int_field"},
+         {"sort":"int_field desc"}),
+    ),
     HighlightOptions:(
         ({"fields":"int_field"},
          {"hl":True, "hl.fl":"int_field"}),
@@ -303,12 +322,26 @@ good_option_data = {
         ({"fields":"text_field", "count":1},
          {"mlt":True, "mlt.fl":"text_field", "mlt.count":1}),
         ),
+    FieldLimitOptions:(
+        ({},
+         {}),
+        ({"fields":"int_field"},
+         {"fl":"int_field"}),
+        ({"fields":["int_field", "text_field"]},
+         {"fl":"int_field,text_field"}),
+        ({"score": True},
+         {"fl":"score"}),
+        ({"all_fields": True, "score": True},
+         {"fl":"*,score"}),
+        ({"fields":"int_field", "score": True},
+         {"fl":"int_field,score"}),
+        ),
     }
 
 def check_good_option_data(OptionClass, kwargs, output):
     optioner = OptionClass(schema)
     optioner.update(**kwargs)
-    assert optioner.options == output
+    assert optioner.options() == output
 
 def test_good_option_data():
     for OptionClass, option_data in good_option_data.items():
@@ -329,6 +362,10 @@ bad_option_data = {
         {"sort":"yes"}, # invalid choice
         {"offset":-1}, # invalid value
         ),
+    SortOptions:(
+        {"field":"myarse"}, # Undefined field
+        {"field":"string_field"}, # Multivalued field
+        ),
     HighlightOptions:(
         {"fields":"myarse"}, # Undefined field
         {"oops":True}, # undefined option
@@ -345,9 +382,9 @@ bad_option_data = {
     }
 
 def check_bad_option_data(OptionClass, kwargs):
-    paginate = OptionClass(schema)
+    option = OptionClass(schema)
     try:
-        paginate.update(**kwargs)
+        option.update(**kwargs)
     except SolrError:
         pass
     else:
@@ -359,21 +396,124 @@ def test_bad_option_data():
             yield check_bad_option_data, OptionClass, kwargs
 
 
-class TestAndOrSyntax(object):
-    # I'm not completely sure of the syntax for this yet ...
-    def setUp(self):
-        solr_search = SolrSearch(interface)
-        self.q = solr_search.query()
+complex_boolean_queries = (
+    (lambda q: q.query("hello world").filter(q.Q(text_field="tow") | q.Q(boolean_field=False, int_field__gt=3)),
+     [('fq', u'text_field:tow OR (boolean_field:false AND int_field:{3 TO *})'), ('q', u'hello\\ world')]),
+    (lambda q: q.query("hello world").filter(q.Q(text_field="tow") & q.Q(boolean_field=False, int_field__gt=3)),
+     [('fq', u'boolean_field:false AND text_field:tow AND int_field:{3 TO *}'), ('q',  u'hello\\ world')]),
+# Test various combinations of NOTs at the top level.
+# Sometimes we need to do the *:* trick, sometimes not.
+    (lambda q: q.query(~q.Q("hello world")),
+     [('q',  u'NOT hello\\ world')]),
+    (lambda q: q.query(~q.Q("hello world") & ~q.Q(int_field=3)),
+     [('q',  u'NOT hello\\ world AND NOT int_field:3')]),
+    (lambda q: q.query("hello world", ~q.Q(int_field=3)),
+     [('q', u'hello\\ world AND NOT int_field:3')]),
+    (lambda q: q.query("abc", q.Q("def"), ~q.Q(int_field=3)),
+     [('q', u'abc AND def AND NOT int_field:3')]),
+    (lambda q: q.query("abc", q.Q("def") & ~q.Q(int_field=3)),
+     [('q', u'abc AND def AND NOT int_field:3')]),
+    (lambda q: q.query("abc", q.Q("def") | ~q.Q(int_field=3)),
+     [('q', u'abc AND (def OR (*:* AND NOT int_field:3))')]),
+    (lambda q: q.query(q.Q("abc") | ~q.Q("def")),
+     [('q', u'abc OR (*:* AND NOT def)')]),
+    (lambda q: q.query(q.Q("abc") | q.Q(~q.Q("def"))),
+     [('q', u'abc OR (*:* AND NOT def)')]),
+# Make sure that ANDs are flattened
+    (lambda q: q.query("def", q.Q("abc"), q.Q(q.Q("xyz"))),
+     [('q', u'abc AND def AND xyz')]),
+# Make sure that ORs are flattened
+    (lambda q: q.query(q.Q("def") | q.Q(q.Q("xyz"))),
+     [('q', u'def OR xyz')]),
+# Make sure that empty queries are discarded in ANDs
+    (lambda q: q.query("def", q.Q("abc"), q.Q(), q.Q(q.Q() & q.Q("xyz"))),
+     [('q', u'abc AND def AND xyz')]),
+# Make sure that empty queries are discarded in ORs
+    (lambda q: q.query(q.Q() | q.Q("def") | q.Q(q.Q() | q.Q("xyz"))),
+     [('q', u'def OR xyz')]),
+# Test cancellation of NOTs.
+    (lambda q: q.query(~q.Q(~q.Q("def"))),
+     [('q', u'def')]),
+    (lambda q: q.query(~q.Q(~q.Q(~q.Q("def")))),
+     [('q', u'NOT def')]),
+# Test it works through sub-sub-queries
+    (lambda q: q.query(~q.Q(q.Q(q.Q(~q.Q(~q.Q("def")))))),
+     [('q', u'NOT def')]),
+# Even with empty queries in there
+    (lambda q: q.query(~q.Q(q.Q(q.Q() & q.Q(q.Q() | ~q.Q(~q.Q("def")))))),
+     [('q', u'NOT def')]),
+# Test escaping of AND, OR, NOT
+    (lambda q: q.query("AND", "OR", "NOT"),
+     [('q', u'"AND" AND "NOT" AND "OR"')]),
+# Test exclude (rather than explicit NOT
+    (lambda q: q.query("blah").exclude(q.Q("abc") | q.Q("def") | q.Q("ghi")),
+     [('q', u'blah AND NOT (abc OR def OR ghi)')]),
+# Try boosts
+    (lambda q: q.query("blah").query(q.Q("def")**1.5),
+     [('q', u'blah AND def^1.5')]),
+    (lambda q: q.query("blah").query((q.Q("def") | q.Q("ghi"))**1.5),
+     [('q', u'blah AND (def OR ghi)^1.5')]),
+    (lambda q: q.query("blah").query(q.Q("def", ~q.Q("pqr") | q.Q("mno"))**1.5),
+     [('q', u'blah AND (def AND ((*:* AND NOT pqr) OR mno))^1.5')]),
+# And boost_relevancy
+    (lambda q: q.query("blah").boost_relevancy(1.5, int_field=3),
+     [('q', u'blah OR (blah AND int_field:3^1.5)')]),
+    (lambda q: q.query("blah").query("blah2").boost_relevancy(1.5, int_field=3),
+     [('q', u'(blah AND blah2) OR (blah AND blah2 AND int_field:3^1.5)')]),
+# And ranges
+    (lambda q: q.query(int_field__any=True),
+     [('q', u'int_field:[* TO *]')]),
+    (lambda q: q.query("blah", ~q.Q(int_field__any=True)),
+     [('q', u'blah AND NOT int_field:[* TO *]')]),
+)
 
-    def test_or_example(self):
-        Q = self.q.Q
-        self.q.query("hello world").filter(Q(text_field="tow") | Q(boolean_field=False, int_field__gt=3))
-        assert self.q.execute() == \
-            {'q': u'"hello world"', 'fq': u'(text_field:tow) OR (boolean_field:false int_field:{3 TO *})'}
+def check_complex_boolean_query(solr_search, query, output):
+    p = query(solr_search).params()
+    try:
+        assert p == output
+    except AssertionError:
+        if debug:
+            print p
+            print output
+            import pdb;pdb.set_trace()
+            raise
+        else:
+            raise
+    # And check no mutation of the base object
+    q = query(solr_search).params()
+    try:
+        assert p == q
+    except AssertionError:
+        if debug:
+            print p
+            print q
+            import pdb;pdb.set_trace()
+            raise
 
-    def test_and_example(self):
-        Q = self.q.Q
-        self.q.query("hello world").filter(Q(text_field="tow") & Q(boolean_field=False, int_field__gt=3))
-        assert self.q.execute() == \
-            {'q': u'"hello world"', 'fq': u'(text_field:tow) AND (boolean_field:false int_field:{3 TO *})'}
+def test_complex_boolean_queries():
+    solr_search = SolrSearch(interface)
+    for query, output in complex_boolean_queries:
+        yield check_complex_boolean_query, solr_search, query, output
 
+
+param_encode_data = (
+    ({"int":3, "string":"string", "unicode":u"unicode"},
+     [("int", "3"), ("string", "string"), ("unicode", "unicode")]),
+    ({"int":3, "string":"string", "unicode":u"\N{UMBRELLA}nicode"},
+     [("int", "3"), ("string", "string"), ("unicode", "\xe2\x98\x82nicode")]),
+    ({"int":3, "string":"string", u"\N{UMBRELLA}nicode":u"\N{UMBRELLA}nicode"},
+     [("int", "3"), ("string", "string"), ("\xe2\x98\x82nicode", "\xe2\x98\x82nicode")]),
+    ({"true":True, "false":False},
+     [("false", "false"), ("true", "true")]),
+    ({"list":["first", "second", "third"]},
+     [("list", "first"), ("list", "second"), ("list", "third")]),
+)
+
+def check_url_encode_data(kwargs, output):
+    # Convert for pre-2.6.5 python
+    s_kwargs = dict((k.encode('utf8'), v) for k, v in kwargs.items())
+    assert params_from_dict(**s_kwargs) == output
+
+def test_url_encode_data():
+    for kwargs, output in param_encode_data:
+        yield check_url_encode_data, kwargs, output
